ActorBehavior subclass: #ConsumerPrinter
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Actalk-Ex-BoundedBuffer'!
ConsumerPrinter comment:
'Class ConsumerPrinter displays the actual consumed items on the transcript window. It acts as the reply destination of asynchronous getAndReplyTo: requests.'!


!ConsumerPrinter methodsFor: 'script'!

reply: item
	"Trace the actual consumption on the Transcript window."

	Transcript show: 'Consumes: ' , item printString; cr! !

SingleAcceptActorBehavior subclass: #Producer
	instanceVariableNames: 'buffer delay '
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Actalk-Ex-BoundedBuffer'!
Producer comment:
'Class Producer implements the producer of a bounded buffer.

Instance Variables:

	buffer	<BoundedBuffer actor>	a bounded buffer on which it will issue put requests.
	delay	<Integer>				the amount of seconds between requests.

Note that class Producer is defined as a subclass of class SingleAcceptActorBehavior in order to avoid wasting infinite processes. This is because class Producer (as well as class Consumer) is used only once and with only one run message when running a bounded buffer example.

Method runPut: runs the producer to issue a certain amount of put: (asynchronous) requests to the buffer.'!


!Producer methodsFor: 'initialize'!

buffer: aBoundedBuffer delay: seconds

	buffer := aBoundedBuffer.
	delay := seconds! !

!Producer methodsFor: 'script'!

runPut: max
	"Run max times a put: request to the bounded buffer."

	1 to: max do: [:i |
		Transcript show: 'Produces: ' , i printString; cr.
		buffer put: i.
		(Delay forSeconds: delay) wait]! !

ActorBehavior subclass: #BoundedBuffer
	instanceVariableNames: 'contents maxSize '
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Actalk-Ex-BoundedBuffer'!
BoundedBuffer comment:
'Class BoundedBuffer implements the seminal example of synchronization.

The point is following: put: requests should not be accepted while the buffer is full. Conversely get requests should not be accepted while the buffer is empty. The acceptance of the methods is depending on the state of the buffer, that is put: method should be disabled (in other words, put: messages should not be accepted) when it is full, and get method should be disabled when it is empty.

This is a seminal example of synchronization on state conditions. Here we don''t deal with intra-object concurrency where we would have to also control possible concurrent activations. See a first discussion of this point in class EmptyBoundedBuffer comment, in category Actalk-Replace-Examples. See examples of bounded buffer classes dealing with intra-object concurrency and activation synchronization within categories Actalk-Synchro-Counters-Ex and Actalk-Synchro-ASC-Ex.

In this example actual implementation of the bounded buffer relies on an internal bounded buffer object (class BoundedBufferObject) in order to focus on synchronization issues and not on data representation. In extensions with intra-object concurrency (e.g., category Actalk-Synchro-Counters-Ex) we will use an alternative implementation using an array and two indexes to clearly separate put and get accesses (accessing to distinct zones, a put and a get may take place concurrently).

In this first implementation (this example will be revisited in all kernel extensions dealing with synchronization), we face the absence of synchronization constructs because there is no way of controlling message acceptance. We can only acknowledge the failure in case of accepting disabled methods.

Instance Variables:

	contents		<BoundedBufferObject>	the bounded buffer object.
	maxSize	<Integer>				its maximum size, that is the maximum number of items.

The bounded buffer is shared by a producer (class Producer) and a consumer (class Consumer). It is possible to adjust relative speed of producer and consumer in order to check fullness or emptyness of the bounded buffer.
For instance using a ratio greater than 1 (e.g., 4 as in first example comment) will make the consumer work slower and as a consequence the bounded buffer will become full. A ratio lower than 1 (e.g., 1/4 as in second example comment) will make the bounded buffer become empty.

In all variations of class BoundedBuffer it may be specially useful to trace events to show the ordering of requests reception, acceptance and completion. You may set traces on a general class (for all example classes, evaluate ActorBehavior setTrace: true) or more specific ones (ex: BoundedBuffer setTrace: true).

Note that the general issue is to be able to control the scheduling (starting of activation) of messages computations. Motivations may be to preserve consistency of the state (as for this BoundedBuffer example), avoid deadlock (see example of dining philosophers in category Actalk-Ext-Pool-Examples), avoid starvation, optimize services (for instance a printer would print out small jobs in priority), etc... By scheduling, we don''t mean that activations have to be serialized as in the case of objects with no intra-object concurrency.
The general problem is to find strategies to express such conditions to be general and expressive enough, modular enough to be easily inherited, and easy to implement efficiently. This general problem is still an open problem as we need to find best compromises/tradeoffs.'!


!BoundedBuffer methodsFor: 'initialize'!

initialize: anInteger

	maxSize := anInteger.
	contents := BoundedBufferObject new maxSize: maxSize! !

!BoundedBuffer methodsFor: 'script'!

getAndReplyTo: r

	| item |
	contents isEmpty						"If empty,"
		ifTrue: [r reply: #getFailed]			"then, return a failure acknowledgement."
		ifFalse:							"otherwise,"
			[item := contents removeFirst.	"remove and return the first item."
			r reply: item]!

put: item

	contents isFull						"If full,"
		ifTrue: [Print reply: #putFailed]		"then, print a failure acknowledgement."
		ifFalse: [contents addLast: item	]	"Otherwise, add the item at the end."! !

!BoundedBuffer methodsFor: 'printing'!

printOn: aStream
	"Print as its class name, then its contents."

	super printOn: aStream.
	contents printOn: aStream! !
"-- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- "!

BoundedBuffer class
	instanceVariableNames: ''!


!BoundedBuffer class methodsFor: 'instance creation'!

new: maxSize

	^self new initialize: maxSize! !

!BoundedBuffer class methodsFor: 'example'!

exampleSize: maxSize numberItems: numberItems speedRatio: ratio
	"BoundedBuffer exampleSize: 3 numberItems: 8 speedRatio: 4"
	"BoundedBuffer exampleSize: 3 numberItems: 8 speedRatio: 1/4"

	| buffer producer consumer |
	buffer := (self new: maxSize) actor.
	producer := (Producer new buffer: buffer delay: 1) actor.
	consumer := (Consumer new buffer: buffer delay: ratio) actor.
	producer runPut: numberItems.
	consumer runGet: numberItems replyTo: ConsumerPrinter new actor! !

BoundedBuffer subclass: #BoundedBuffer2
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Actalk-Ex-BoundedBuffer'!
BoundedBuffer2 comment:
'Class BoundedBuffer2 is a first step towards management of synchronization.
It is defined as a subclass of class BoundedBuffer.

Rather than discarding methods which are disabled (e.g., put: method while the buffer is full), the buffer resends such requests to itself. This assumes that the unbounded (possibly infinite) cycle of reaccepting/resending the message will be eventually stopped (triggered by another request, e.g., a get request so the buffer could stop being full). This uses the method resend: (of class Basic2Actor) which transparently ensures current behavior process yields way to others (because reaccept/resend infinite loop would keep the processor and forbid other actors, like consumer to send other requests).

Notice that this strategy of resending messages breaks down the assumption of transmission message ordering. Some message which initially arrived before another one may finally be accepted after it because of the constant reordering within the mailbox. This may be a problem for some algorithms (e.g., class PrimeFilter) on which we would then need to add some timestamping strategy in order to recover the initial ordering at sending time. This may also be specified by synchronization constraints. See category Actalk-Synchro-Activation as an example.
Such resending of message is also used for class NullBoundedBuffer (category Actalk-Ext-Replace-Examples) as well as for several classes to express synchronization. (To list them, find users of methods resend: and resendWithoutYield: defined in class Basic2Actor).'!


!BoundedBuffer2 methodsFor: 'script'!

getAndReplyTo: r

	| item |
	contents isEmpty						"If empty,"
		ifTrue: [aself resend:				"then, resend the original message..."
					(Message selector: #getAndReplyTo: argument: r)]	"...reconstructed."
		ifFalse:							"otherwise,"
			[item := contents removeFirst.	"remove and return the first item."
			r reply: item]!

put: item

	contents isFull						"If full,"
		ifTrue: [aself resend:				"then, resend the original message..."
					(Message selector: #put: argument: item)]	"...reconstructed."
		ifFalse: [contents addLast: item	]	"Otherwise, add the item at the end."! !
"-- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- "!

BoundedBuffer2 class
	instanceVariableNames: ''!


!BoundedBuffer2 class methodsFor: 'example'!

exampleSize: maxSize numberItems: numberItems speedRatio: ratio
	"BoundedBuffer2 exampleSize: 3 numberItems: 8 speedRatio: 4"
	"BoundedBuffer2 exampleSize: 3 numberItems: 8 speedRatio: 1/4"

	super exampleSize: maxSize numberItems: numberItems speedRatio: ratio! !

SingleAcceptActorBehavior subclass: #Consumer
	instanceVariableNames: 'buffer delay '
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Actalk-Ex-BoundedBuffer'!
Consumer comment:
'Class Consumer implements the consumer of a bounded buffer.

Instance Variables:

	buffer	<BoundedBuffer actor>	a bounded buffer on which it will issue get requests.
	delay	<Integer>				the amount of seconds between requests.

Note that class Consumer is defined as a subclass of class SingleAcceptActorBehavior in order to avoid wasting infinite processes. This is because class Consumer (as well as class Producer) is used only once and with only one run message when running a bounded buffer example.

There are several methods to run the consumer, depending on the type of the communication protocols (asynchronous, synchronous, implicit reply) with the bounded buffer actor.

Method runGet:replyTo: runs the consumer to issue a certain amount of asynchronous getAndReplyTo: requests to the buffer. (A ConsumerPrinter actor is specified as reply destination).
Method runNowGet: runs the consumer to issue a certain amount of synchronous (now type) getAndReplyTo: requests to the buffer, assumed to be an Abcl type actor. (See category Actalk-Ext-Abcl).
Method runGet:  runs the consumer to issue a certain amount of synchronous get requests to the buffer, assumed to be an implicit reply type actor. (See category Actalk-Ext-ImplicitReply).

Method runGet2: runs alternatively get and get2 requests.
Method runGget: runs alternatively get and gget requests. (Both methods get2 returning two elements at once, and gget to be disabled after a put, are introduced in synchronization extension classes, see Actalk-Synchro-* categories). They have been proposed as seminal examples of inheritance anomalies by Satoshi Matsuoka and Akinori Yonezawa in the book Research Directions in COOP, MIT-Press, 1993.'!


!Consumer methodsFor: 'initialize'!

buffer: aBoundedBuffer delay: seconds

	buffer := aBoundedBuffer.
	delay := seconds! !

!Consumer methodsFor: 'script'!

runGet2: max
	"Run max/2 times a get2 request to the bounded buffer (to remove max items).
	Assume the bounded buffer is an implicit reply actor."

	| items |
	max / 2 timesRepeat:
		[items := buffer get2 value.
		Transcript show: 'Consumes: ' , items printString; cr.
		(Delay forSeconds: delay) wait]!

runGet: max
	"Run max times a get (synchronous) request to the bounded buffer.
	Assume the bounded buffer is an implicit reply actor."

	| item |
	max timesRepeat:
		[Transcript show: 'Consumer requests'; cr.
		item := buffer get value.	"Wait on the value."
		Transcript show: 'Consumes: ' , item printString; cr.
		(Delay forSeconds: delay) wait]!

runGet: max replyTo: r
	"Run max times a getAndReplyTo: (asynchronous) request to the bounded buffer."

	max timesRepeat:
		[Transcript show: 'Consumer requests'; cr.
		buffer getAndReplyTo: r.
		(Delay forSeconds: delay) wait]!

runGetRear: max
	"Run max times alternatively get and getRear requests to the bounded buffer.
	Assume the bounded buffer is an implicit reply actor."

	| item isGetRear |
	isGetRear := true.
	max timesRepeat:
		[isGetRear		"Flip flop."
			ifTrue: [item := buffer getRear]
			ifFalse: [item := buffer get].
		isGetRear := isGetRear not.
		"We fork the wait on the reply in order to keep sending get and getRear requests.
		Because the point is to see how get and getRear could be computed concurrently."
		[Transcript show: 'Consumes: ' , item value printString; cr] fork.
		Transcript show: 'Consumes: ' , item printString; cr.
		(Delay forSeconds: delay) wait	]!

runGget: max
	"Run max times alternatively get and gget requests to the bounded buffer.
	Assume the bounded buffer is an implicit reply actor."

	| item isGget |
	isGget := true.
	max timesRepeat:
		[isGget		"Flip flop."
			ifTrue: [item := buffer gget]		"This is a future, not the value yet."
			ifFalse: [item := buffer get].
		isGget := isGget not.
		"We fork the wait on the reply to keep sending get and gget requests.
		Otherwise we get (immediately) stuck on the (first) gget request coming after a put: request."
		[Transcript show: 'Consumes: ' , item value printString; cr] fork.
		(Delay forSeconds: delay) wait	]!

runNowGet: max
	"Run max times a get (synchronous) request to the bounded buffer.
	Assume the bounded buffer is an Abcl type actor."

	| item |
	max timesRepeat:
		[Transcript show: 'Consumer requests'; cr.
		item := buffer getAndReplyTo: #now.	"Wait on the result."
		Transcript show: 'Consumes: ' , item printString; cr.
		(Delay forSeconds: delay) wait]! !

OrderedCollection variableSubclass: #BoundedBufferObject
	instanceVariableNames: 'maxSize '
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Actalk-Ex-BoundedBuffer'!
BoundedBufferObject comment:
'Class BoundedBufferObject is a simple implementation of a bounded buffer object as an ordered collection.

Instance Variables:

	maxSize	<Integer>	the maximum size of the bounded buffer
							(that is the maximum number of items it may contain).

Method addLast: is equivalent to a put:, and method removeFirst is equivalent to a get.
Method addLast: is redefined to check if the buffer is full. This is an internal check to trap whenever further synchronization constructs or programming is deficient.
State predicates (methods isFull and isEmpty) are also defined for convenience.'!


!BoundedBufferObject methodsFor: 'accessing'!

maxSize: anInteger

	maxSize := anInteger! !

!BoundedBufferObject methodsFor: 'adding'!

addLast: item
	"Catch and trigger error if buffer is full."
	"This provides (some) checking of synchronization constructs, implementation, and specification."

	self size >= maxSize
		ifTrue: [self error: 'buffer is full']
		ifFalse: [super addLast: item]! !

!BoundedBufferObject methodsFor: 'state predicates'!

isEmpty

	^self size = 0!

isFull

	^self size = maxSize! !
