BodyAcceptActorBehavior subclass: #PoolActorBehavior
	instanceVariableNames: 'postBlock '
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Actalk-Ext-Pool'!
PoolActorBehavior comment:
'Classes PoolActor and PoolActorBehavior implement Pool body model, explicit message acceptance, synchronous message passing, and (object-like) public routines.

Class PoolActorBehavior is defined as a subclass of class BodyAcceptActorBehavior.

It defines a default body which accepts any kind of message successively (as for standard actor behaviors).
Class PoolActorBehavior also provides post actions which can be executed after returning the value, thus freeing the sender from unneeded waiting.
Class PoolActorBehavior also provide public routines. Public routines offer object-like methods which are directly computed by the behavior.
They are useful to encapsulate public protocols. In fact they can be necessary to avoid deadlock as it is the case when there is recursion (possibly mutual). (Remember that all message passing in POOL is synchronous thus recursion leads to deadlock). See for instance the example of class SymbolTable.
Public routines (not to be confused with private routines, see class ActorBehavior comment) are tagged by their protocol (public routines). In order to have a fast dispatch when receiving a message, the set of public routines selectors is precomputed (see class method computePublicRoutines).

Instance Variables:

	postBlock	<nil/BlockClosure>	the block specifying post actions of current method
									(nil if no post actions are specified by current method).

This instance variable acts as a container and a flag to ensure that post actions are completed before accepting next message.

Acknowledgements:
The implementation of POOL in Actalk was initially designed by a group of students of DESS of University of Nantes, under the guidance of Jean Bezivin and Olivier Roux. This first implementation has since been revised, further extended, and decomposed in the AcceptActorBehavior, BodyAcceptActorBehavior, and PoolActorBehavior hierarchy.'!


!PoolActorBehavior methodsFor: 'body'!

body
	"Default body is to answer/accept all successive messages."
	"In other words, default body behavior is equivalent to standard actor behavior process."

	[true] whileTrue:
		[self acceptNextMessage]! !

!PoolActorBehavior methodsFor: 'public routines check'!

isPublicRoutineMessage: aMessage
	"Check if the message is a public routine selector."

	^self publicRoutines includes: aMessage selector!

publicRoutines
	"Specify and return the array of public routine selectors."
	"Default set is empty."

	^#()! !

!PoolActorBehavior methodsFor: 'message handling'!

acceptMessage: aMessage
	"Check if a post action has been specified.
	If the case, compute it after completing the method (and before accepting next one)."

	postBlock := nil.
	super acceptMessage: aMessage.
	postBlock isNil ifFalse:
		[postBlock value]!

acceptPublicRoutineMessage: aMessage
	"Perform the public routine message as in standard Smalltalk."
	"Notice that it bypasses the performMessage: method
	(because it is redefined in class ImplicitReplyActorBehavior
	in order to set the returned value into the extended message).
	In case of public routine handling we stick to standard message handling."

	^self perform: aMessage selector withArguments: aMessage arguments! !

!PoolActorBehavior methodsFor: 'post actions'!

post: aBlock
	"Specify post actions (specified within a block) which are to be computed
	after completing the method and its return.
	This frees the sender from unneeded waiting, as post actions will take place after return.
	Note that meanwhile post actions must be computed BEFORE accepting next message."
	"Typical use is for reinitialization."

	postBlock := aBlock! !

!PoolActorBehavior methodsFor: 'actor creation'!

actorClass

	^PoolActor! !
"-- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- "!

PoolActorBehavior class
	instanceVariableNames: ''!


!PoolActorBehavior class methodsFor: 'initialize'!

initialize
	"Initialize the set of public routine selectors."
	"Needs to be recomputed if one adds new public routine methods."

	self computePublicRoutines! !

!PoolActorBehavior class methodsFor: 'public routines'!

computePublicRoutines
	"Compute the set of public routine selectors by collecting them along the hierarchy.
	Create the method publicRoutines accordingly.
	Do nothing if no public routine is defined within the behavior
	(then inheritance of method publicRoutines will prevail)."

	| publicRoutines superclassPublicRoutines instance |
	publicRoutines := self selectorsInCategory: #'public routines'.
	publicRoutines isEmpty
		ifFalse:
			["Compute the set of public routines of the behavior superclass."
			instance := self superclass someInstance.		"We need one of its instances,"
			instance isNil								"possibly by creating one."
				ifTrue: [instance := self superclass new].
			superclassPublicRoutines := instance publicRoutines.
			"Now compute the new union set and the new method."
			self compile: 'publicRoutines\\	^' withCRs ,
						(superclassPublicRoutines , publicRoutines) printString
				classified: #'public routines check'.
			"Recompute the public routines in the subclasses if there are some."
			self subclasses do:
				[:subclass | subclass computePublicRoutines]]! !

IRActor subclass: #PoolActor
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Actalk-Ext-Pool'!
PoolActor comment:
'Classes PoolActor and PoolActorBehavior implement POOL body model, explicit message acceptance, synchronous message passing, and (object-like) public routines.
See an article onto the POOL language by Pierre America in the OOCP book.
Class PoolActor implements synchronous message passing as a standard and dispatches onto public routine handling.
Class PoolActor is defined as a subclass of class IRActor (to handle implicit reply).
See class PoolActorBehavior comment.'!


!PoolActor methodsFor: 'message passing'!

doesNotUnderstand: aMessage
	"First check if the message is requesting a public routine.		
	If so, the behavior computes directly the message.
	Otherwise, implicitly wait onto the future returned (to implement synchronous message passing)."

	^(behavior isPublicRoutineMessage: aMessage)
		ifTrue: [behavior acceptPublicRoutineMessage: aMessage]
		ifFalse: [(self futureSend: aMessage) value]! !
PoolActorBehavior initialize!

