SharedQueue subclass: #MailBox
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Actalk-Kernel-Basic'!
MailBox comment:
'Class MailBox is a subclass of standard class SharedQueue.
SharedQueue provides basic functionalities of a mailbox to receive and retrieve messages (plus mutual exclusion and suspension while the mailbox is empty).
Class MailBox implements some accessing methods specific to Actalk (looking for specific patterns and incrementally look for a message). See the protocol message accesing. Note that these are in someway low-level implementation routines.'!


!MailBox methodsFor: 'message accessing'!

atIndex: i
	"Look for and fetch the ith message in the mailbox.
	Suspend until it exists if not yet (by waiting for new messages as many times as needed)."
	"Useful for class GuardsSynchroActorBehavior."
	"WARNING: this method is safe if only one behavior at a time uses it.
	It is unsafe for concurrent use by many behaviors
	(because it makes assumption on the monotony of the mailbox)."
	"The manipulation of readSynch semaphore is somewhat tricky.
	We need to because we want to suspend until a new message comes
	BUT we don't want to decrement the number of items in the queue
	(until we actually find and remove a matching message)."
	"Used by class GuardsSynchroActorBehavior."

	| numberOfWaits |		"We need to record number of waits on readSynch semaphore"
	numberOfWaits := 0.		"in order to restore them afterwards."
		"Because this number represents the number of messages in the mailbox."
	[[accessProtect critical: [i > contents size]] valueUninterruptably]
		whileTrue:
			[readSynch wait.		"Wait for next incoming message."
			numberOfWaits := numberOfWaits + 1].
	numberOfWaits timesRepeat: [readSynch signal].	"Restore the semaphore excessive signals."
	^contents at: i!

checkFirstMessageAmong: collectionOfSelectors
	"Look for the first message matching one of the selectors.
	As opposed to method firstMessageAmong: which is blocking,
	unBlockingLookForFirstMessageAmong: is not.
	If a matching message is found it removes and returns it. Otherwise it returns nil."
	"May be useful for class AcceptActorBehavior."

	^self internalFirstMessageAmong: collectionOfSelectors withReadSynchWait: true!

firstMessageAmong: collectionOfSelectors
	"Look for the first message matching one of the selectors.
	It is blocking and suspend until it finds any matching one.
	(This means it suspends onto incoming messages as long as necessary)."
	"The manipulation of readSynch semaphore excessive signals is somewhat tricky.
	We need to because we want to suspend until a new message comes
	BUT we don't want to decrement the number of items in the queue
	(until we actually find and remove a matching message)."
	"Used by classes AbclActorBehavior, AcceptActorBehavior and StateSynchroActorBehavior."

	| message numberOfWaits |
		"First instant look for a matching message."
	message := self internalFirstMessageAmong: collectionOfSelectors withReadSynchWait: true.
		"We need to record number of waits on readSynch semaphore in order to restore them afterwards."
		"Because this number represents the number of messages in the mailbox."
	numberOfWaits := 0.	
	[message isNil]		"While no matching message has been found yet,"
		whileTrue:		"we need to check again repeatedly after each new incoming message."
			[readSynch wait.		"Wait for next incoming message."
			numberOfWaits := numberOfWaits + 1.
				"The flag/argument of withReadSynchWait: is false, because we already waited for current message."
			message := self internalFirstMessageAmong: collectionOfSelectors withReadSynchWait: false].
	numberOfWaits - 1 timesRepeat: [readSynch signal].	"Restore the semaphore excessive signals."
	^message										"(minus one, because one for the removed item)."!

internalFirstMessageAmong: collectionOfSelectors withReadSynchWait: isWait
	"Look for the first message matching one of the selectors.
	It is not blocking and returns nil if no matching message is found."
	"The isWait flag tells if a wait onto the readSynch semaphore should be produced
	in case a matching message has been found and will be removed.
	Producing a wait is the usual case to tell that one item has been removed.
	However in some case (call from the loop of method firstMessageAmong:)
	the wait has already been produced, and doing it again would suspend computation."

	| message |
	[accessProtect critical:
		[message := contents
					detect: [:msg | collectionOfSelectors includes: msg selector]
					ifNone: [].
		message isNil				"If found, we remove (and then return) it."
			ifFalse:
				[isWait ifTrue: [readSynch wait].	"We know we don't have to wait, but we want to tell we remove one item."
				contents remove: message]]] valueUninterruptably.
	^message!

removeAtIndex: i
	"Remove the ith message of the mailbox."
	"Useful for class GuardsSynchroActorBehavior."
	"WARNING: this method is safe if only one behavior uses it.
	It is unsafe for concurrent use by many behaviors
	(because it makes assumption on the monotony of the mailbox)."

	readSynch wait.		"We assume that we don't have to wait, but we want to tell we remove one item."
	^[accessProtect critical: [contents removeAtIndex: i]] valueUninterruptably! !

!MailBox methodsFor: 'printing'!

printOn: aStream
	"Print as its contents (an OrderedCollection)."

	contents printOn: aStream! !

Object subclass: #Basic1ActorBehavior
	instanceVariableNames: 'aself '
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Actalk-Kernel-Basic'!
Basic1ActorBehavior comment:
'Class Basic1ActorBehavior is the class which implements the behavior of an actor.
The behavior is the object which eventually consumes incoming messages and processes them.

Basic kernel classes such as Basic1ActorBehavior are basic classes.
The entry point in the Actalk kernel is the set of the two kernel classes Actor and ActorBehavior.

See more insights into how to define and use actors in class ActorBehavior comment.

Instance Variables:

	aself	<Actor>	is the actual actor owning this behavior.

aself is useful to send recursive messages (asynchronously) to the actor (through its mailbox).
Pseudo-variable self can be still used and will imply a standard recursive message passing (directly) to the behavior.

Method initializeAndSetAself: initializes the behavior.
Method actalkPrivateInitialize (also provided by class Basic1Actor) is a generic initialize method. Default does nothing.
Method setBehavior: sets its process by calling method behaviorProcessBlock which dispatches to the specified process behavior (default is accepting serially all pending messages).
Method acceptNextMessage decomposes into two generic methods: acceptMessage: and nextMessage (default is performing next message from the mailbox).
Method actorClass returns the default class for actors (here Basic1ActorBehavior).

Acknowledgements: Loic Lescaudron initially designed the basic kernel library (classes Basic1Actor(Behavior) and Basic2Actor(Behavior).'!


!Basic1ActorBehavior methodsFor: 'initialize'!

actalkPrivateInitialize
	"Default initialize: do nothing."
	"It is not named initialize to ensure there is no name clash/shadowing with user (re)definition of initialize.
	Redefinition of class method new ensures automatic calling of method actalkPrivateInitialize."!

initializeAndSetAself: anActor
	"Initialize the behavior, i.e. its reference to its actor and set its autonomous process."

	aself := anActor.		"Cross reference onto the actor."
	self setProcess		"Create and start the process implementing the autonomy of the behavior."! !

!Basic1ActorBehavior methodsFor: 'process setting'!

acceptSeriallySuccessiveMessagesBehaviorProcessBlock
	"Endless loop to serially consume successive pending messages."

	^[[true] whileTrue: [self acceptNextMessage]]!

behaviorProcessBlock
	"Create and return the process block which provides its autonomy to the behavior."
	"The only purpose of this method is to dispatch onto various possible process block behaviors."
	"Default behavior is an endless loop to consume pending messages."

	^self acceptSeriallySuccessiveMessagesBehaviorProcessBlock!

setProcess
	"Create and start the behavior autonomous process."
	"We use a generic method to specify the process block (and not the process)
	so we may change the kind of process to specific (e.g., time-slicing) kinds of process if needed."

	self behaviorProcessBlock newProcess resume! !

!Basic1ActorBehavior methodsFor: 'message handling'!

acceptMessage: aMessage
	"Accept a message."
	"Default behavior is simply computing (perform) it."

	self performMessage: aMessage!

acceptNextMessage
	"Accepting next message."
	"(It gets suspended until there is one, see method nextMessage)."

	self acceptMessage: self nextMessage!

nextMessage
	"Next message is the next one taken from the mailbox."
	"Note that this fetching is suspended until the mailbox is not empty,
	thanks to the SharedQueue standard class."

	^aself mailBox next! !

!Basic1ActorBehavior methodsFor: 'actor creation'!

actor
	"Create an actor, once the behavior is specified."
	"The class of the actor (e.g., class Actor) is specified and returned by method actorClass."
	"This is similar to method defaultControllerClass of class View."

	^self actorClass newBehavior: self!

actorClass
	"Specify the associated class of actors, by default class Actor."
	"Is redefined when defining new pairs of classes *ActorBehavior and *Actor."

	^Basic1Actor! !
"-- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- "!

Basic1ActorBehavior class
	instanceVariableNames: ''!


!Basic1ActorBehavior class methodsFor: 'instance creation'!

new

	^super new actalkPrivateInitialize! !

Basic1ActorBehavior subclass: #Basic2ActorBehavior
	instanceVariableNames: 'process '
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Actalk-Kernel-Basic'!
Basic2ActorBehavior comment:
'Class Basic2ActorBehavior (paired with class Basic2Actor) offers a few more facilities than basic kernel classes Basic1Actor and Basic1ActorBehavior.

Class Basic2ActorBehavior provides a handle onto the background process which allow check, control (suspend, resume and terminate). The terminate action will be useful to do some cleanup in order to recover background processes, see ActorBehavior class methods classCleanUp and cleanUp.

It defines a new basic behavior process (block) in the process patterns library. This second one (the first and default process pattern accepts serially all incoming messages) accepts only one message. It will prove useful when implementing actors with multiple behaviors (see class ReplaceActorBehavior).

Class Basic2ActorBehavior also provides three generic event handling methods which are associated to actor events (receiving a message, accepting it, and completing its activation) which may be redefined for many purposes (trace, schedule, step...). These methods are respectively named eventReceive:, eventAccept:, and eventComplete:. They take the message on which event takes place as an argument. Note that these methods are user-oriented and are mostly used to trace events onto the transcript window (see class ActorBehavior class methods to automatically set up such tracing facilities). Kernel event methods also exist (and by default just dispatch onto user event methods). This is because generic events will be used to implement some synchronization mechanisms (like in class CountersSynchroActorBehavior). Therefore we prefer to avoid name clashes between user events and system events.

Last, it provides a checking when creating an actor (method actor) that the behavior is not shared with another actor. (This would be catastrophic because it would break the synchronization protection).

Class method allScriptSelectors returns all script selectors defined within a behavior class hierarchy. This method will be specially useful for synchronization behavior classes.

Basic kernel classes such as Basic2ActorBehavior are basic classes.
The entry point in the Actalk kernel is the set of the two kernel classes Actor and ActorBehavior.

Instance Variables:

	process		<Process>	is the process implementing the autonomy of the behavior.'!


!Basic2ActorBehavior methodsFor: 'accessing'!

process
	"Accessing to the process of the behavior.
	May be useful for external control."

	^process! !

!Basic2ActorBehavior methodsFor: 'process setting'!

acceptOneMessageBehaviorProcessBlock
	"Consume next pending message."
	"Will be useful for classes of actors with multiple behaviors, e.g., class ReplaceActorBehavior."

	^[self acceptNextMessage]!

setProcess
	"Provide a handle to the process."

	process := self behaviorProcessBlock newProcess.
	process resume! !

!Basic2ActorBehavior methodsFor: 'process control'!

resume
	"Resume the process of the behavior.
	Assume that the process has been suspended."

	process resume!

suspend
	"Suspend the process of the (behavior) of the actor.
	Assume there is a process running."

	process suspend!

terminate
	"Terminate the process of the behavior.
	Useful for cleanup (reclaiming resources)."

	process isNil
		ifFalse: [process terminateNoSignal]! !

!Basic2ActorBehavior methodsFor: 'message handling'!

acceptMessage: aMessage
	"Add the triggering of generic event methods for accepting and completing computation of a message."

	self kernelEventAccept: aMessage.	
	super acceptMessage: aMessage.
	self kernelEventComplete: aMessage! !

!Basic2ActorBehavior methodsFor: 'events'!

eventAccept: aMessage
	"This method is called when the behavior accepts a message."
	"This method may be redefined in subclasses."
	"Default behavior is empty."

	"Below are possible examples of use to:
trace:
	Transcript show: self printString , ' accept (' , aMessage compactPrintString , ')'; cr.
schedule:
	Processor yield.
broadcast change to dependents (e.g., useful for MVC interface):
	self changed: #eventAccept with: aMessage.
step:
	self halt.
	"!

eventComplete: aMessage
	"This method is called once the behavior completed computation of a message."
	"This method may be redefined in subclasses."
	"Default behavior is empty."

	"Below are possible examples of use to:
trace:
	Transcript show: self printString , ' complete (' , aMessage compactPrintString , ')'; cr.
schedule:
	Processor yield.
broadcast change to dependents (e.g., useful for MVC interface):
	self changed: #eventComplete with: aMessage.
step:
	self halt.
	"!

eventReceive: aMessage
	"This method is called once the behavior has received a new message."
	"This method may be redefined in subclasses."
	"Default behavior is empty."

	"Below are possible examples of use to:
trace:
	Transcript show: self printString , ' receive (' , aMessage compactPrintString , ')'; cr.
schedule:
	Processor yield.
broadcast change to dependents (e.g., useful for MVC interface):
	self changed: #eventReceive with: aMessage.
step:
	self halt.
	"!

kernelEventAccept: aMessage
	"This method is called when the behavior accept a message."
	"This is the basic kernel method which should not be subclassed by application classes."

	self eventAccept: aMessage!

kernelEventComplete: aMessage
	"This method is called once the behavior completed computation of a message."
	"This is the basic kernel method which should not be subclassed by application classes."

	self eventComplete: aMessage!

kernelEventReceive: aMessage
	"This method is called after the behavior has received a new message."
	"This is the basic kernel method which should not be subclassed by application classes."

	self eventReceive: aMessage! !

!Basic2ActorBehavior methodsFor: 'actor creation'!

actor
	"Check if the behavior has already been used to create an actor."

	^aself isNil
		ifTrue: [super actor]
		ifFalse: [self error: 'is already the behavior of an actor']!

actorClass

	^Basic2Actor! !
"-- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- "!

Basic2ActorBehavior class
	instanceVariableNames: ''!


!Basic2ActorBehavior class methodsFor: 'accessing method dictionary'!

allScriptSelectors
	"Answer the set of all the script selectors known by a behavior class."

	^self allSelectorsInCategory: #script untilClass: ActorBehavior! !

MinimalObject subclass: #Basic1Actor
	instanceVariableNames: 'mailBox behavior '
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Actalk-Kernel-Basic'!
Basic1Actor comment:
'Class Basic1Actor is the class which implements the external shell of an actor (seen from the outside, and refered to by other actors).
It acts mostly as a container for the mailbox which will contain incoming messages and the behavior which will process them. (See class Basic1ActorBehavior comment).
Its main purpose is also to implement asynchronous message passing.

Basic kernel classes such as Basic1ActorBehavior are basic classes.
The entry point in the Actalk kernel is the set of the two kernel classes Actor and ActorBehavior.

Instance Variables:

	mailBox		<MailBox>		contains incoming messages.
	behavior	<ActorBehavior>	is the actual behavior of the actor, that is the object which will actually consume and process incoming messages.

Redefined error handling method doesNotUnderstand: calls method asynchronousSend: to send an asynchronous message.
Method setBehavior: sets and initializes the behavior of an actor.
Method actalkPrivateInitialize (also provided by class Basic1ActorBehavior) is a generic initialize method. It is automatically called when creating instances by redefinition of class method new. It is not named initialize to ensure there is no name clash/shadowing with possible user (re)definition of initialize. It creates a new mailbox for the actor.

Acknowledgement:
The organization of Actalk kernel into a first basic kernel has been introduced by Loic Lescaudron within the development of the Actalk full environment (including extended MVC and generic time-slicing scheduler) implemented in Smalltalk-80 version 2.5.'!


!Basic1Actor methodsFor: 'initialize'!

actalkPrivateInitialize
	"Initialize the mailbox, by default a MailBox (SharedQueue)."
	"It is not named initialize to ensure there is no name clash/shadowing with user (re)definition of initialize.
	Redefinition of class method new ensures automatic calling of method actalkPrivateInitialize."

	mailBox := MailBox new! !

!Basic1Actor methodsFor: 'accessing'!

mailBox
	"This accessor method is necessary for the behavior(s) to access to the mailbox of the actor."

	^mailBox! !

!Basic1Actor methodsFor: 'behavior setting'!

setBehavior: aBehavior
	"Set (assign) a behavior to the actor and initialize it."

	behavior := aBehavior.
	behavior initializeAndSetAself: self		"Initialize the behavior and set its cross reference to the actor."! !

!Basic1Actor methodsFor: 'message passing'!

asynchronousSend: aMessage
	"Asynchronous message sending equals the deposit of the message in the mailbox of the actor (at the end of it)."

	mailBox nextPut: aMessage!

doesNotUnderstand: aMessage
	"A message sent to an Actalk actor will trigger an error and be resent in the asynchronous way."
	"Assumption of triggering an error (being unknown) is ensured by class MinimalObject."

	self asynchronousSend: aMessage! !
"-- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- "!

Basic1Actor class
	instanceVariableNames: ''!


!Basic1Actor class methodsFor: 'instance creation'!

new

	^super new actalkPrivateInitialize!

newBehavior: aBehavior
	"Create and initialize a new actor with the specified behavior."

	^self new setBehavior: aBehavior! !

Basic1Actor subclass: #Basic2Actor
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Actalk-Kernel-Basic'!
Basic2Actor comment:
'Class Basic2Actor (paired with class Basic2ActorBehavior) implements a few more facilities than basic kernel class Basic1Actor (and Basic1ActorBehavior).
The pair provides access to the behavior process, this is useful for control and cleanup. It also offers generic events associated to actor events of receiving, accepting, and completing a message invocation. See class Basic2ActorBehavior comment.
Class Basic2Actor also provides direct acces by method asObject to the behavior (thus violating its encapsulation and synchronization, but may be useful in some situations, see method comment).'!


!Basic2Actor methodsFor: 'accessing'!

asObject
	"Return the behavior of an actor."
	"This bypasses the actor shell and thus considers an actor as an almost standard Smalltalk object.
	This could be used to consult the contents of the behavior.
	This facility should be used extremely cautiously because it clashes with the protection/serialization of the actor.
	Consequently it should be used to consult (read) but not to change (write) the state of the behavior.
	Examples of use are when building user interfaces onto actors."

	^behavior! !

!Basic2Actor methodsFor: 'message passing'!

asynchronousSend: aMessage
	"Add the triggering of generic event after receiving a message."
	"Note that the method eventArrive: will also be defined in class UndefinedObject
	so that if there is no behavior yet (for instance in case of FutureActors) nil does not signal an error."

	super asynchronousSend: aMessage.
	behavior kernelEventReceive: aMessage!

resend: aMessage
	"Resend the message by enqueueing it to the mailbox.
	Useful to implement (naive) delay of a message when the conditions of its acceptance are not yet fullfilled.
	(See for instance naive implementation of synchronization in class BoundedBuffer2, category Actalk-Ex-BoundedBuffer).

	WARNING: This may result in an infinite accept/resend cycle. (See below).

	Also notice that it does not call the method asynchronousSend:,
	this is because we don't want to trigger the receive event again."

	self mailBox nextPut: aMessage.
	Processor yield		"Give way to other processes. This is NECESSARY.
						Otherwise there is strong chance this actor ALWAYS keep the processor.
						(By infinitely accepting and resending the same message)."!

resendWithoutYield: aMessage
	"Resend the message by enqueueing it to the mailbox.
	Useful to implement (naive) delay of a message when the conditions of its acceptance are not yet fullfilled.

	WARNING: This MAY result in an infinite accept/resend cycle.
	No processor yield is performed. (NO PROTECTION!!).
	However this is useful when a yield is unneeded in case of no accept loop
	(this is the case for subprocesses computing only one message).
	(See for instance naive implementation of synchronization in method doesNotUnderstand:
	of class EmptyBoundedBuffer within category Actalk-Ext-Replace-Examples.
	Also see class CountersActorBehavior within category Actalk-Synchro-Counters.)

	Also notice that it does not call the method asynchronousSend:,
	this is because we don't want to trigger the receive event again."

	self mailBox nextPut: aMessage! !

!Basic2Actor methodsFor: 'reset'!

reset
	"Full reset of an actor, that is, its mailbox and its process."

	self initialize.
	self resetProcess! !

!Basic2Actor methodsFor: 'process control'!

resetProcess
	"Reset the process of (the behavior) of an actor."

	behavior setProcess!

resume
	"Resume the process of the (behavior) of the actor."

	behavior resume!

suspend
	"Suspend the process of the (behavior) of the actor."

	behavior suspend!

terminate
	"Terminate the process of the (behavior) of the actor.
	Useful for cleanup (reclaiming resources)."

	behavior terminate! !

!Basic2Actor methodsFor: 'printing'!

printOn: aStream
	"External representation of an actor: & followed by the behavior. Ex: &a Counter."

	aStream nextPut: $&.
	behavior printOn: aStream! !
