ConcurrentActorBehavior subclass: #SuspendActorBehavior
	instanceVariableNames: 'suspendedProcessList controlSemaphore '
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Actalk-Synchro-Suspend'!
SuspendActorBehavior comment:
'Class SuspendActorBehavior implements suspension of methods.
Within a method body, one may express suspension on some condition.
This is analog to ConcurrentSmalltalk-II relinquish: method.
This is also analog to Kessel monitors except that monitors synchronize external processes, as here processes are internally created for every starting request.

We need to be able to suspend some method computation, and then start another one or restart one previously suspended (and whose suspension condition became false).
Therefore we need a main control behavior process. It will act as a scheduler of message/method activations. It launchs suprocesses for computing requests. It gets then suspended but gets resumed once a method computation is suspended or completed.
The main control behavior process is following.
It first checks if one previously suspended method subprocess is ready to resume (by checking their conditions). If the case, it resumes it. Otherwise it creates a new behavior subprocess to process next pending message. (This is a lazy creation as it first waits for a pending message if needed). Then (in both cases) it suspends until next method completion or suspension.

Class SuspendActorBehavior is defined as a subclass of class ConcurrentActorBehavior as it holds subprocesses.

Instance Variables:

	suspendedProcessList 	<OrderedCollection>	holds the ordered collection of waiting (suspended) subprocesses.
							Each item of the collection is an association:
								<condition block , semaphore on which the process waits>.

	controlSemaphore		<Semaphore>	is the main control semaphore to synchronize the main central behavior process
							and the starting/suspension/resumption/completion of subprocesses created to compute messages/methods.'!


!SuspendActorBehavior methodsFor: 'initialize'!

actalkPrivateInitialize

	super actalkPrivateInitialize.
	suspendedProcessList := OrderedCollection new.
	controlSemaphore := Semaphore new! !

!SuspendActorBehavior methodsFor: 'message handling'!

acceptNextMessage
	"Accepting next message means here first checking for a suspended method to resume.
	Otherwise we create a new subprocess for next pending message.
	Then current main control behavior process gets suspended
	(until next completion or suspension) before reentering this method."

	"association = <condition block , semaphore on which the process waits>."
	| anAssociation |	
	"First look for a suspended subprocess whose suspension condition has become true."
	(suspendedProcessList isEmpty		"If no suspended subprocess is waiting,"
		or:	[anAssociation := suspendedProcessList detect: [:association | association key value]
												ifNone: [].
			anAssociation isNil])			"or if none is ready to resume,"
		ifTrue:
			"then, start a new subprocess to accept next message."
			[self setSubProcessFor: self nextMessage]
		ifFalse:
			"Otherwise remove the suspended subprocess and resume it
			(by signaling its suspending semaphore)."
			[suspendedProcessList remove: anAssociation.	
			anAssociation value signal].
		"In both cases, the main control behavior waits for resumption
		(after next method completion or suspension)."
	controlSemaphore wait! !

!SuspendActorBehavior methodsFor: 'subprocess setting'!

behaviorSubProcessBlockFor: aMessage
	"This is the behavior process block for a subprocess computing a new message.
	Accept and consume a message.
	Then resume the main control behavior process afterwards."

	^[self subProcessAcceptMessage: aMessage.
	   controlSemaphore signal]! !

!SuspendActorBehavior methodsFor: 'process control'!

terminate
	"Terminate all suspended subprocesses at first."

	suspendedProcessList do:
		[:association | association value terminateProcess].
	super terminate! !

!SuspendActorBehavior methodsFor: 'suspension'!

waitUntil: conditionBlock
	"Suspends current method computation on a condition (block)."
	"Equivalent to relinquish: construct in ConcurrentSmalltalk-II."

	| aSemaphore |		
	conditionBlock value ifFalse:		"If condition is true do nothing, otherwise,"
		[aSemaphore := Semaphore new.	"Create a new semaphore,"
		suspendedProcessList addLast:	"add current subprocess info to the suspension list,"
		"association = <condition block , semaphore on which the process waits>."
			(Association new key: conditionBlock value: aSemaphore).
		controlSemaphore signal.			"Then, first resume the control process,"
		aSemaphore wait]				"and then suspend the subprocess onto the semaphore."

		"Note that when signaled, the suspended subprocess resumes here,
		and immediately returns to resumption of processing the method body."! !
